/*
 * Galatea Dialog Manager:
 * (c)2006 Takuya NISHIMOTO (nishi@hil.t.u-tokyo.ac.jp)
 *
 * $Id: ISystemEventLogger.java,v 1.2 2006/12/11 01:34:26 nishi Exp $
 */
package galatea.dialog;

public interface ISystemEventLogger {
	public void outputVoiceStarted(String string);
	public void inputEventReceived(String xmlevent);
	public void documentLocationChanged(String filename);
	public void dialogStateChanged(String string);
	public void srcFileUpdated(String srcAsString);
	public void translatedFileUpdated(String translatedAsString);
	public void grammarFileUpdated(String grammar);
	public void fatalError(String error);
}
